# -*- coding: utf-8 -*-
#
# New BSD license
#
# Copyright (c) DR0ID
# This file is part of HG_pyknic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL DR0ID BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
TODO: module description


Versioning scheme based on: http://en.wikipedia.org/wiki/Versioning#Designating_development_stage

::

      +-- api change, probably incompatible with older versions
      |     +-- enhancements but no api change
      |     |
    major.minor[.build[.revision]]
                   |        |
                   |        +-|* x for x bugfixes
                   |
                   +-|* 0 for alpha (status)
                     |* 1 for beta (status)
                     |* 2 for release candidate
                     |* 3 for (public) release

.. versionchanged:: 0.0.0.0
    initial version



Layers:

bird            b                ---
street light    l2           -----------------
tree            t                  -------------------
moped light     l1       ---------------------
street          s     ---------------------------------------
street behind   g     ---------------------------------------

^ normal blit
+ blend add
* blend mult
no operator precedence except the brackets, just left to right
a = ambient color (darkness, assume nearly black)

in spritesystem this was used:
disadvantage: no sprites between layers of light, all layers are illuminated
pixel = (a + l1 + l2) * (g ^ s ^ t ^ b)

can only have max the original color
pixel = (g ^ s) * (a + l1 + l2) ^ (t * (a + l2)) ^ (b * a)

this has color compression!
pixel = (g ^ s) * a + (l1 + l2) ^ (t * a + l2) ^ (b * a)


Instructions:

escape: quits, works also while pressing the window close button
up, down, left right: Use arrow keys to move objects around.
1-9: All are selectable through numbers 1 - 9 (1 lowest layer...)
'a': Ambient can be selected by pressing 'a' and using up/down arrows.
'r': reset to initial values (positions, ambient, etc.)
F1-F4: toggle algorithms, to verify that one algorithm has no effect on another (because sharing of sprites!)


"""
from __future__ import print_function
import pygame
import time

__version__ = '1.0.0.0'

# for easy comparison as in sys.version_info but digits only
__version_info__ = tuple([int(d) for d in __version__.split('.')])

__author__ = "DR0ID"
__email__ = "dr0iddr0id {at} gmail [dot] com"
__copyright__ = "DR0ID @ 2015"
__credits__ = ["DR0ID"]  # list of contributors
__maintainer__ = "DR0ID"
__license__ = "New BSD license"

__all__ = []  # list of public visible parts of this module 



class Sprite(object):
    def __init__(self, img_tuple, x=0, y=0, layer=0, is_light=False):
        image, name = img_tuple
        self.image = image
        self.name = name
        self.rect = image.get_rect() if image else pygame.Rect(0, 0, 0, 0)
        self.rect.topleft = int(x), int(y)
        self.is_light = is_light
        self.layer = layer


def main():

    screen_width = 1024
    screen_height = 900
    screen_size = screen_width, screen_height
    screen_background_color = (255, 0, 255)
    tile_background_color = (255, 0, 255, 255)
    text_color = (255, 255, 255)
    text_background_color = (0, 0, 0)
    tile_border_color = (255, 255, 255)
    object_high_light_color = (255, 255, 0)

    pygame.init()
    # screen = pygame.display.set_mode(screen_size, pygame.RESIZABLE | pygame.SRCALPHA, 32)
    screen = pygame.display.set_mode(screen_size, 0, 32)

    step = 10
    x_count = 3
    y_count = 3
    tile = pygame.Surface((screen_width // x_count, screen_height // y_count), pygame.SRCALPHA)
    tile.fill(tile_background_color)
    tile_rect = tile.get_rect()
    font = pygame.font.Font(None, 20)
    numbers = []
    for i in range(x_count * y_count):
        numbers.append(font.render(str(i + 1), 0, text_color, text_background_color))

    # load sprites
    sprites, ambient_sprite, ambient = init(tile.get_size())
    assert len(sprites) < 11
    selected = sprites[2]
    highlight_selected = True
    show_help = True
    help_image = pygame.image.load("help.png")

    is_tile_visible = [True] * x_count * y_count
    is_tile_visible[0] = False
    is_tile_visible[2] = False

    draw_methods = [draw1, draw2, draw3, draw4, draw5, draw6, draw7, draw8, draw9]
    assert len(draw_methods) >= x_count * y_count

    running = True
    while running:
        # events
        # for event in pygame.event.get():
        for event in [pygame.event.wait()]:

            if event.type == pygame.QUIT:
                running = False
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    running = False
                elif event.key == pygame.K_RIGHT:
                    selected.rect.x += step
                elif event.key == pygame.K_LEFT:
                    selected.rect.x -= step
                elif event.key == pygame.K_UP:
                    selected.rect.y -= step
                elif event.key == pygame.K_DOWN:
                    selected.rect.y += step
                elif event.key in (
                        pygame.K_1, pygame.K_2, pygame.K_3, pygame.K_4, pygame.K_5, pygame.K_6, pygame.K_7, pygame.K_8,
                        pygame.K_9):
                    index = (event.key - pygame.K_1) % len(sprites)
                    selected = sprites[index]
                elif event.key == pygame.K_a:
                    selected = ambient_sprite
                elif event.key == pygame.K_r:
                    sprites, ambient_sprite, ambient = init(tile.get_size())
                    selected = sprites[2]
                elif event.key in (
                        pygame.K_F1, pygame.K_F2, pygame.K_F3, pygame.K_F4, pygame.K_F5, pygame.K_F6, pygame.K_F7,
                        pygame.K_F8, pygame.K_F9, pygame.K_F10):
                    index = (event.key - pygame.K_F1) % len(is_tile_visible)
                    is_tile_visible[index] = not is_tile_visible[index]
                elif event.key == pygame.K_SPACE:
                    highlight_selected = not highlight_selected
                elif event.key == pygame.K_TAB:
                    show_help = not show_help

                a_value = ambient_sprite.rect.y
                a_value %= 255
                ambient = (a_value,) * 3
                print(selected.rect, ambient)

        # draw
        for x in range(x_count):
            for y in range(y_count):
                idx = y + x * y_count
                if is_tile_visible[idx]:
                    screen_tile = tile.copy()

                    # use blit algorithm and take the time it takes
                    pg_start = pygame.time.get_ticks()
                    # start = time.time()
                    draw_methods[idx](screen_tile, ambient, sprites)
                    # duration = (time.time() - start) * 1000.0
                    duration = pygame.time.get_ticks() - pg_start

                    # draw highlight of object
                    if highlight_selected:
                        # rect
                        r = selected.rect.clip(screen_tile.get_rect())
                        pygame.draw.rect(screen_tile, object_high_light_color, r, 2)

                        # name
                        name_surf = font.render(selected.name, 0, text_color, text_background_color)
                        screen.blit(name_surf, (2, 30))

                    # draw white border
                    pygame.draw.rect(screen_tile, tile_border_color, tile_rect, 1)

                    # number of tile and duration
                    screen_tile.blit(numbers[idx], (2, 2))
                    duration_surf = font.render(str(duration) + " ms", 0, text_color, text_background_color)
                    screen_tile.blit(duration_surf, (22, 2))

                    # blit tile
                    pos = (x * screen_width // x_count, y * screen_height // y_count)
                    screen.blit(screen_tile, pos)

        if show_help:
            help_rect = help_image.get_rect(center=(screen_width // 2, screen_height // 2))
            screen.blit(help_image, help_rect)

        pygame.display.flip()
        screen.fill(screen_background_color)

    pygame.quit()


def init(size):
    # street = Sprite(pygame.image.load("street.png"), -15-170, -370, layer=0)
    # tomato = Sprite(pygame.image.load("tomato.png"), 150-100, 90, layer=1)
    # front_light = Sprite(pygame.image.load("frontlight.png"), 140-100, 120, is_light=True, layer=2)
    # tree = Sprite(pygame.image.load("leaves.png"), 110, 90, layer=5)
    # street_light = Sprite(pygame.image.load("streetlight.png"), 190-100, 80, is_light=True, layer=10)
    # bird = Sprite(pygame.image.load("bird.png"), 260-100, 120, layer=11)
    #
    # red_light = Sprite(pygame.image.load("red.png"), 60, 20, is_light=True, layer=2)
    # green_light = Sprite(pygame.image.load("green.png"), 10, 50, is_light=True, layer=2)
    # blue_light = Sprite(pygame.image.load("blue.png"), 0, 0, is_light=True, layer=2)

    w, h = size
    print(w, h)  # 340, 265   190 = w * x -> x = 190/

    def _y(y):
        return int(float(y) / 265.0 * h)

    def _x(x):
        return int(float(x) / 340.0 * w)

    def load_img(name):
        img = pygame.image.load(name)
        return img, name

    street = Sprite(load_img("street.png"), _x(0), _y(0), layer=0)
    tomato = Sprite(load_img("tomato.png"), _x(190), _y(90), layer=1)
    front_light = Sprite(load_img("frontlight.png"), _x(190), _y(80), is_light=True, layer=2)
    tree = Sprite(load_img("leaves.png"), _x(70), _y(90), layer=5)
    street_light = Sprite(load_img("streetlight.png"), _x(170), _y(80), is_light=True, layer=10)
    bird = Sprite(load_img("bird.png"), _x(170), _y(120), layer=11)

    red_light = Sprite(load_img("red.png"), _x(260), _y(-10), is_light=True, layer=2)
    green_light = Sprite(load_img("green.png"), _x(260), _y(20), is_light=True, layer=2)
    blue_light = Sprite(load_img("blue.png"), _x(230), _y(0), is_light=True, layer=2)

    ambient_sprite = Sprite((None, "ambient"), -1, 15)
    ambient = (ambient_sprite.rect.y,) * 3
    return [street, tomato, front_light, tree, street_light, bird, red_light, green_light,
            blue_light], ambient_sprite, ambient


def draw1(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = (a + l1 + l2) * (g ^ s ^ t ^ b)
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    a = screen.copy()
    a.fill(ambient)
    a.blit(front_light.image, front_light.rect, None, pygame.BLEND_ADD)
    a.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    a.blit(red_light.image, red_light.rect, None, pygame.BLEND_ADD)
    a.blit(green_light.image, green_light.rect, None, pygame.BLEND_ADD)
    a.blit(blue_light.image, blue_light.rect, None, pygame.BLEND_ADD)

    screen.blit(street.image, street.rect)
    screen.blit(tomato.image, tomato.rect)
    screen.blit(tree.image, tree.rect)
    screen.blit(bird.image, bird.rect)
    screen.blit(a, (0, 0), None, pygame.BLEND_MULT)


def draw2(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = (g ^ s) * (a + l1 + l2) ^ (t * (a + l2)) ^ (b * a)
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    a = screen.copy()
    a.fill(ambient)

    screen.blit(street.image, street.rect)
    screen.blit(tomato.image, tomato.rect)

    a1 = a.copy()
    a1.blit(front_light.image, front_light.rect, None, pygame.BLEND_ADD)
    a1.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    a1.blit(red_light.image, red_light.rect, None, pygame.BLEND_ADD)
    a1.blit(green_light.image, green_light.rect, None, pygame.BLEND_ADD)
    a1.blit(blue_light.image, blue_light.rect, None, pygame.BLEND_ADD)

    screen.blit(a1, (0, 0), None, pygame.BLEND_MULT)

    a2 = a.copy()
    a2.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    t = tree.image.copy()
    t.blit(a2, (-tree.rect.left, -tree.rect.top), None, pygame.BLEND_MULT)

    screen.blit(t, tree.rect)

    b = bird.image.copy()
    b.fill(ambient, None, pygame.BLEND_MULT)

    screen.blit(b, bird.rect)


def draw3(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = (g ^ s) * a + (l1 + l2) ^ (t * a + l2) ^ (b * a)
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites

    a = screen.copy()
    a.fill(ambient)

    screen.blit(street.image, street.rect)
    screen.blit(tomato.image, tomato.rect)
    screen.blit(a, (0, 0), None, pygame.BLEND_MULT)

    l = screen.copy()
    l.fill((0, 0, 0))
    l.blit(front_light.image, front_light.rect, None, pygame.BLEND_ADD)
    l.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    l.blit(red_light.image, red_light.rect, None, pygame.BLEND_ADD)
    l.blit(green_light.image, green_light.rect, None, pygame.BLEND_ADD)
    l.blit(blue_light.image, blue_light.rect, None, pygame.BLEND_ADD)

    screen.blit(l, (0, 0), None, pygame.BLEND_ADD)

    t = tree.image.copy()
    t.fill(ambient, None, pygame.BLEND_MULT)
    tx = street_light.rect.left - tree.rect.left
    ty = street_light.rect.top - tree.rect.top
    t.blit(street_light.image, (tx, ty), None, pygame.BLEND_ADD)

    screen.blit(t, tree.rect)

    b = bird.image.copy()
    b.fill(ambient, None, pygame.BLEND_MULT)

    screen.blit(b, bird.rect)

def draw4(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = (g ^ s) * (a + l1 + l2) + 0.5*l1 + 0.5*l2 ^ (t * (a + l2) + 0.5*l2) ^ (b * a)
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    a = screen.copy()
    a.fill(ambient)

    l1 = front_light.image.copy()
    factor = (int(0.5 * 255),) * 3
    l1.fill(factor, None, pygame.BLEND_MULT)
    l2 = street_light.image.copy()
    l2.fill(factor, None, pygame.BLEND_MULT)
    r = red_light.image.copy()
    r.fill(factor, None, pygame.BLEND_MULT)
    g = green_light.image.copy()
    g.fill(factor, None, pygame.BLEND_MULT)
    b = blue_light.image.copy()
    b.fill(factor, None, pygame.BLEND_MULT)

    screen.blit(street.image, street.rect)
    screen.blit(tomato.image, tomato.rect)

    a1 = a.copy()
    a1.blit(front_light.image, front_light.rect, None, pygame.BLEND_ADD)
    a1.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    a1.blit(red_light.image, red_light.rect, None, pygame.BLEND_ADD)
    a1.blit(green_light.image, green_light.rect, None, pygame.BLEND_ADD)
    a1.blit(blue_light.image, blue_light.rect, None, pygame.BLEND_ADD)

    screen.blit(a1, (0, 0), None, pygame.BLEND_MULT)

    screen.blit(l1, front_light.rect, None, pygame.BLEND_ADD)
    screen.blit(l2, street_light.rect, None, pygame.BLEND_ADD)
    screen.blit(r, red_light.rect, None, pygame.BLEND_ADD)
    screen.blit(g, green_light.rect, None, pygame.BLEND_ADD)
    screen.blit(b, blue_light.rect, None, pygame.BLEND_ADD)

    a2 = a.copy()
    a2.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    t = tree.image.copy()
    t.blit(a2, (-tree.rect.left, -tree.rect.top), None, pygame.BLEND_MULT)
    t.blit(l2, street_light.rect.move(-tree.rect.left, -tree.rect.top), None, pygame.BLEND_ADD)

    screen.blit(t, tree.rect)

    b = bird.image.copy()
    b.fill(ambient, None, pygame.BLEND_MULT)

    screen.blit(b, bird.rect)

def draw5(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = (g ^ s) * (a + l1 + l2) + (1-a)*0.5*l1 + (1-a)*0.5*0.5*l2 ^ (t * (a + l2) + (1-a)*0.5*0.5*l2) ^ (b * a)
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    a = screen.copy()
    a.fill(ambient)

    amb_factor = (255.0 - ambient[0]) / 255.0
    factor = (int(amb_factor * 0.5 * 255),) * 3
    factor2 = (int(amb_factor * 0.5 * 255 * 0.5),) * 3  # need to divide by number of layer, otherwise n times!
    l1 = front_light.image.copy()
    l1.fill(factor, None, pygame.BLEND_MULT)
    l2 = street_light.image.copy()
    l2.fill(factor2, None, pygame.BLEND_MULT)
    r = red_light.image.copy()
    r.fill(factor, None, pygame.BLEND_MULT)
    g = green_light.image.copy()
    g.fill(factor, None, pygame.BLEND_MULT)
    b = blue_light.image.copy()
    b.fill(factor, None, pygame.BLEND_MULT)

    screen.blit(street.image, street.rect)
    screen.blit(tomato.image, tomato.rect)

    a1 = a.copy()
    a1.blit(front_light.image, front_light.rect, None, pygame.BLEND_ADD)
    a1.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    a1.blit(red_light.image, red_light.rect, None, pygame.BLEND_ADD)
    a1.blit(green_light.image, green_light.rect, None, pygame.BLEND_ADD)
    a1.blit(blue_light.image, blue_light.rect, None, pygame.BLEND_ADD)

    screen.blit(a1, (0, 0), None, pygame.BLEND_MULT)

    # spotlight glimmer
    screen.blit(l1, front_light.rect, None, pygame.BLEND_ADD)
    screen.blit(l2, street_light.rect, None, pygame.BLEND_ADD)
    screen.blit(r, red_light.rect, None, pygame.BLEND_ADD)
    screen.blit(g, green_light.rect, None, pygame.BLEND_ADD)
    screen.blit(b, blue_light.rect, None, pygame.BLEND_ADD)

    a2 = a.copy()
    a2.blit(street_light.image, street_light.rect, None, pygame.BLEND_ADD)
    t = tree.image.copy()
    t.blit(a2, (-tree.rect.left, -tree.rect.top), None, pygame.BLEND_MULT)
    t.blit(l2, street_light.rect.move(-tree.rect.left, -tree.rect.top), None, pygame.BLEND_ADD)

    screen.blit(t, tree.rect)

    b = bird.image.copy()
    b.fill(ambient, None, pygame.BLEND_MULT)

    screen.blit(b, bird.rect)

def draw6(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = (g ^ s) * (a + l1 + l2) ^ (t * (a + l2)) ^ (b * a)
    """
    # street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    sprites = sorted(sprites, key=lambda s: s.layer)
    lights = [light for light in sprites if light.is_light]

    a1 = screen.copy()
    a1.fill(ambient)

    amb_factor = (255.0 - ambient[0]) / 255.0
    factor = (int(amb_factor * 0.5 * 255),) * 3
    # factor = (int(0.5 * 255),) * 3

    l_layer = sprites[0].layer
    light_idx = 0
    s = pygame.Surface(screen.get_size(), pygame.SRCALPHA)
    fill_color = (255, 255, 255, 0)
    s.fill(fill_color)  # transparent surface

    for spr in sprites:
        if spr.is_light:

            if l_layer != spr.layer:
                a1.fill(ambient)
                for i in lights[light_idx:]:
                    a1.blit(i.image, i.rect, None, pygame.BLEND_RGB_ADD)

                s.blit(a1, (0, 0), None, pygame.BLEND_RGBA_MULT)

                # for i in lights[light_idx:]:
                #     j = i.image.copy()
                #     j.fill(factor, None, pygame.BLEND_RGB_MULT)
                #     s.blit(j, i.rect, None, pygame.BLEND_RGB_ADD)

                screen.blit(s, (0, 0))

                s.fill(fill_color)

                l_layer = spr.layer

            light_idx += 1
        else:
            s.blit(spr.image, spr.rect)

    s.fill(ambient, None, pygame.BLEND_RGB_MULT)
    screen.blit(s, (0, 0))

def draw7(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = ?
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    pass

def draw8(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = ?
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    pass

def draw9(screen, ambient, sprites):
    """
    Draws the scene applying following algorithm (see module doc string):

    pixel = ?
    """
    street, tomato, front_light, tree, street_light, bird, red_light, green_light, blue_light = sprites
    pass


if __name__ == '__main__':
    main()
